from __future__ import annotations

import logging
import re
import sys
from typing import Any

from ..core import get_transform, list_transforms
from ..engine.provider_utils import (
    _list_loaded_tensor_names as list_loaded_tensor_names_from_provider,
    list_model_aliases as list_model_aliases_from_provider,
)


logger = logging.getLogger("brainsurgery")


def _collect_completion_candidates(state_dict_provider: Any | None) -> list[str]:
    del state_dict_provider
    commands = list_transforms()
    candidates: set[str] = set()
    for command in commands:
        transform = get_transform(command)
        if not getattr(transform, "completion_requires_payload", True):
            candidates.add(command)
        else:
            candidates.add(f"{command}: ")
    return sorted(candidate for candidate in candidates if candidate)


def _list_model_aliases(state_dict_provider: Any | None) -> set[str]:
    try:
        return list_model_aliases_from_provider(state_dict_provider)
    except Exception:
        logger.debug("Could not list model aliases for completion", exc_info=True)
        return set()


def _list_loaded_tensor_names(state_dict_provider: Any | None) -> dict[str, set[str]]:
    try:
        return list_loaded_tensor_names_from_provider(state_dict_provider)
    except Exception:
        logger.debug("Could not list tensor names for completion", exc_info=True)
        return {}


def _extract_transform_name(line: str) -> str | None:
    stripped = line.strip()
    if stripped.startswith("- "):
        stripped = stripped[2:].lstrip()
    if ":" not in stripped:
        return None
    name = stripped.split(":", 1)[0].strip()
    if not name:
        return None
    if name not in set(list_transforms()):
        return None
    return name


def _infer_active_transform(lines: list[str], line_buffer: str) -> str | None:
    inline = _extract_transform_name(line_buffer)
    if inline:
        return inline
    for line in lines:
        name = _extract_transform_name(line)
        if name:
            return name
    return None


def _collect_payload_candidates(
    *,
    active_transform: str | None,
    state_dict_provider: Any | None,
) -> list[str]:
    candidates: set[str] = {"{ ", "}", "[ ", "]", ", "}

    if active_transform:
        try:
            transform = get_transform(active_transform)
            required_keys = set(getattr(transform, "required_keys", set()) or set())
            allowed_keys = set(getattr(transform, "allowed_keys", set()) or set())
            for key in sorted(required_keys | allowed_keys):
                candidates.add(f"{key}: ")
        except Exception:
            pass

    aliases = _list_model_aliases(state_dict_provider)
    candidates.update(aliases)
    candidates.update(f"{alias}::" for alias in aliases)

    loaded_tensors = _list_loaded_tensor_names(state_dict_provider)
    for alias, names in loaded_tensors.items():
        for name in names:
            candidates.add(name)
            candidates.add(f"{alias}::{name}")

    return sorted(candidate for candidate in candidates if candidate)


def _payload_context(before_cursor: str) -> str:
    trimmed = before_cursor.rstrip()
    if not trimmed:
        return "key"
    if trimmed.endswith("{") or trimmed.endswith(","):
        return "key"

    last_colon = trimmed.rfind(":")
    last_struct_sep = max(trimmed.rfind("{"), trimmed.rfind(","))
    if last_colon > last_struct_sep:
        return "value"
    return "any"


def _current_value_key(before_cursor: str) -> str | None:
    match = re.search(r"([A-Za-z_][A-Za-z0-9_]*)\s*:\s*[^,{}[\]]*$", before_cursor)
    if match is None:
        return None
    return match.group(1)


def _current_value_fragment(before_cursor: str) -> str | None:
    match = re.search(r"[A-Za-z_][A-Za-z0-9_]*\s*:\s*([^,{}[\]]*)$", before_cursor)
    if match is None:
        return None
    return match.group(1)


def _is_committed_value_fragment(raw_fragment: str) -> bool:
    if not raw_fragment:
        return False
    return raw_fragment.endswith(" ") and bool(raw_fragment.strip())


def _is_transform_payload_start(
    *,
    line_buffer: str,
    begidx: int,
    active_transform: str | None,
    endidx: int | None = None,
) -> bool:
    if active_transform is None:
        return False
    if endidx is None:
        endidx = begidx
    if endidx < 0:
        endidx = 0
    if endidx > len(line_buffer):
        endidx = len(line_buffer)
    before = line_buffer[:endidx].strip()
    return before in {f"{active_transform}:", f"{active_transform}: "}


def _match_payload_candidates(
    *,
    text: str,
    line_buffer: str,
    begidx: int,
    endidx: int | None = None,
    payload_candidates: list[str],
    active_transform: str | None = None,
    model_aliases: list[str] | None = None,
) -> list[str]:
    transform = None
    if active_transform is not None:
        try:
            transform = get_transform(active_transform)
        except Exception:
            transform = None

    def _dedupe_preserve_order(items: list[str]) -> list[str]:
        seen: set[str] = set()
        out: list[str] = []
        for item in items:
            if item in seen:
                continue
            seen.add(item)
            out.append(item)
        return out

    def _used_keys(before_cursor: str) -> set[str]:
        return set(re.findall(r"([A-Za-z_][A-Za-z0-9_]*)\s*:", before_cursor))

    def _remaining_key_candidates(before_cursor: str) -> list[str]:
        used_keys = _used_keys(before_cursor)
        remaining = [
            candidate
            for candidate in payload_candidates
            if candidate.endswith(": ") and candidate[:-2] not in used_keys
        ]
        if not remaining:
            return ["}"]
        return [f", {candidate}" for candidate in remaining] + ["}"]

    def _ordered_transform_keys() -> list[str]:
        if transform is None:
            return []
        keys = getattr(transform, "completion_reference_keys", lambda: [])()
        return [key for key in keys if isinstance(key, str) and key]

    def _next_reference_key(before_cursor: str, current_key: str | None) -> str | None:
        if current_key is None:
            return None
        ordered_keys = _ordered_transform_keys()
        if current_key not in ordered_keys:
            return None
        used_keys = _used_keys(before_cursor)
        current_index = ordered_keys.index(current_key)
        for next_key in ordered_keys[current_index + 1 :]:
            if next_key not in used_keys:
                return next_key
        return None

    def _reference_candidates(prefix_text: str, value_key: str | None) -> list[str]:
        alias_prefix_candidates = [
            candidate
            for candidate in payload_candidates
            if candidate.endswith("::") and candidate[:-2] in set(model_aliases or [])
        ]
        if "::" not in prefix_text and len(alias_prefix_candidates) > 1:
            alias_matches = [
                candidate for candidate in alias_prefix_candidates if candidate.startswith(prefix_text)
            ]
            return _dedupe_preserve_order(alias_matches)

        ref_candidates = [
            candidate
            for candidate in payload_candidates
            if (
                not candidate.endswith(": ")
                and candidate not in {"{ ", "}", "[ ", "]", ", "}
            )
        ]
        ref_candidates = [
            candidate for candidate in ref_candidates if not (f"{candidate}::" in ref_candidates)
        ]
        ref_matches = [candidate for candidate in ref_candidates if candidate.startswith(prefix_text)]
        next_key = _next_reference_key(before_cursor, value_key)
        if (
            next_key
            and prefix_text
            and prefix_text in ref_candidates
            and not prefix_text.endswith("::")
        ):
            ref_matches.append(f"{prefix_text}, {next_key}: ")
        return _dedupe_preserve_order(ref_matches)

    def _key_candidates_for_transform(
        *,
        before_cursor: str,
        prefix_text: str,
    ) -> list[str] | None:
        if transform is None:
            return None
        return transform.completion_key_candidates(before_cursor, prefix_text)

    def _value_candidates_for_transform(value_key: str | None, prefix_text: str) -> list[str] | None:
        if transform is None:
            return None
        return transform.completion_value_candidates(value_key, prefix_text, sorted(model_aliases or []))

    def _committed_next_candidates(value_key: str | None) -> list[str] | None:
        if transform is None:
            return None
        return transform.completion_committed_next_candidates(value_key)

    raw_text = text
    prefix = text
    if endidx is None:
        endidx = begidx + len(text)
    if endidx < 0:
        endidx = 0
    if endidx > len(line_buffer):
        endidx = len(line_buffer)
    before_cursor = line_buffer[:endidx]
    ctx = _payload_context(before_cursor)
    if ctx == "key":
        if "," in prefix:
            prefix = prefix.rsplit(",", 1)[-1]
        if "{" in prefix:
            prefix = prefix.rsplit("{", 1)[-1]
        if "}" in prefix:
            prefix = prefix.rsplit("}", 1)[-1]
        prefix = prefix.lstrip()
    if _is_transform_payload_start(
        line_buffer=line_buffer,
        begidx=begidx,
        active_transform=active_transform,
        endidx=endidx,
    ):
        if transform is not None:
            payload_start_candidates = transform.completion_payload_start_candidates(prefix)
            if payload_start_candidates is not None:
                return payload_start_candidates
        if not prefix:
            return ["{ "]
        return [candidate for candidate in ["{ "] if candidate.startswith(prefix)]

    value_key = _current_value_key(before_cursor)
    raw_value_fragment = _current_value_fragment(before_cursor) or ""
    if ctx == "key":
        key_candidates = _key_candidates_for_transform(before_cursor=before_cursor, prefix_text=prefix)
        if key_candidates is not None:
            if raw_text.rstrip().endswith("{"):
                return [f"{raw_text} {candidate}" for candidate in key_candidates if candidate != "}"]
            if before_cursor.rstrip().endswith(",") and not before_cursor.endswith(", "):
                if raw_text.rstrip().endswith(","):
                    return [
                        f"{raw_text} {candidate}" for candidate in key_candidates if candidate != "}"
                    ] + (["}"] if "}" in key_candidates else [])
                return [f" {candidate}" for candidate in key_candidates if candidate != "}"] + (
                    ["}"] if "}" in key_candidates else []
                )
            return key_candidates

    if not prefix:
        if ctx == "key":
            used_keys = _used_keys(before_cursor)
            key_candidates = [
                candidate
                for candidate in payload_candidates
                if (
                    candidate.endswith(": ")
                    and candidate[:-2] not in used_keys
                )
            ]
            if before_cursor.rstrip().endswith(",") and not before_cursor.endswith(", "):
                if raw_text.rstrip().endswith(","):
                    return [f"{raw_text} {candidate}" for candidate in key_candidates]
                return [f" {candidate}" for candidate in key_candidates]
            return key_candidates
        if ctx == "value":
            if _is_committed_value_fragment(raw_value_fragment):
                return _committed_next_candidates(value_key) or _remaining_key_candidates(before_cursor)
            transform_value_candidates = _value_candidates_for_transform(value_key, "")
            if transform_value_candidates is not None:
                return transform_value_candidates
            if value_key in _ordered_transform_keys():
                return _reference_candidates("", value_key)
            return [candidate for candidate in payload_candidates if candidate in {"{ ", "[ ", "]", "}"}]
        return payload_candidates

    if "::" in prefix:
        value_key = _current_value_key(before_cursor)
        if ctx == "value" and _value_candidates_for_transform(value_key, prefix) is not None:
            return _value_candidates_for_transform(value_key, prefix) or []
        if ctx == "value" and value_key in _ordered_transform_keys():
            return _reference_candidates(prefix, value_key)
        return [candidate for candidate in payload_candidates if "::" in candidate and candidate.startswith(prefix)]

    if prefix[0] in "{[]},:":
        return [candidate for candidate in payload_candidates if candidate.startswith(prefix)]

    if prefix.endswith(":") or re.fullmatch(r"[A-Za-z_][A-Za-z0-9_]*:?", prefix):
        if ctx == "key":
            used_keys = _used_keys(before_cursor)
            key_candidates = [
                candidate
                for candidate in payload_candidates
                if (
                    candidate.endswith(": ")
                    and candidate.startswith(prefix)
                    and candidate[:-2] not in used_keys
                )
            ]
            if before_cursor.rstrip().endswith(",") and not before_cursor.endswith(", "):
                if raw_text.rstrip().endswith(","):
                    return [f"{raw_text} {candidate}" for candidate in key_candidates]
                return [f" {candidate}" for candidate in key_candidates]
            return key_candidates

    if ctx == "value":
        value_key = _current_value_key(before_cursor)
        raw_value_fragment = _current_value_fragment(before_cursor) or ""
        if _is_committed_value_fragment(raw_value_fragment):
            committed_next = _committed_next_candidates(value_key) or _remaining_key_candidates(before_cursor)
            return [candidate for candidate in committed_next if candidate.startswith(prefix)]
        transform_value_candidates = _value_candidates_for_transform(value_key, prefix)
        if transform_value_candidates is not None:
            return transform_value_candidates
        if value_key in _ordered_transform_keys():
            return _reference_candidates(prefix, value_key)
        return [
            candidate
            for candidate in payload_candidates
            if not candidate.endswith(": ") and candidate.startswith(prefix)
        ]

    return [candidate for candidate in payload_candidates if candidate.startswith(prefix)]


def _render_completion_preview(matches: list[str], limit: int = 16) -> str:
    if not matches:
        return ""
    shown = matches[:limit]
    remaining = len(matches) - len(shown)
    suffix = f" (+{remaining} more)" if remaining > 0 else ""
    return "  ".join(shown) + suffix


def _completion_display_hook(
    substitution: str,
    matches: list[str],
    longest_match_length: int,
    readline_module: Any | None,
) -> None:
    del substitution, longest_match_length
    preview = _render_completion_preview(matches)
    if not preview:
        return
    sys.stdout.write(f"\nCompletions: {preview}\n")
    try:
        if readline_module is not None:
            readline_module.redisplay()
    except Exception:
        pass


def _configure_readline_completion_bindings(readline_module: Any | None) -> None:
    if readline_module is None:
        return

    commands = [
        "set show-all-if-ambiguous on",
        "set show-all-if-unmodified on",
        "set completion-query-items 200",
        "set page-completions off",
        "set menu-complete-display-prefix on",
        "tab: menu-complete",
        '"\\e[Z": menu-complete-backward',
    ]

    for command in commands:
        try:
            readline_module.parse_and_bind(command)
        except Exception:
            logger.debug("Could not apply readline command %r", command, exc_info=True)


def _is_top_level_completion_position(line_buffer: str, begidx: int) -> bool:
    if begidx > len(line_buffer):
        return False
    before_cursor = line_buffer[:begidx]
    stripped = before_cursor.lstrip()
    if not stripped:
        return True
    if stripped.startswith("- "):
        remainder = stripped[2:]
        return remainder == ""
    return False



from __future__ import annotations

from ..core import UnarySpec, UnaryTransform
from ..core import TensorRef, must_model
from ..core import register_transform
from ..core import StateDictProvider, TransformError


class DeleteTransformError(TransformError):
    pass


class DeleteTransform(UnaryTransform[UnarySpec]):
    name = "delete"
    error_type = DeleteTransformError
    spec_type = UnarySpec
    progress_desc = "Applying delete transforms"
    help_text = (
        "Deletes one or more tensors selected by 'target'.\n"
        "\n"
        "Targets may be specified by name or pattern. The entire tensor is removed; "
        "slicing is not supported.\n"
        "\n"
        "Examples:\n"
        "  delete: { target: ln_f_copy.weight }\n"
        "  delete: { target: '.*_backup' }"
    )

    def apply_to_target(self, spec: UnarySpec, name: str, provider: StateDictProvider) -> None:
        model = must_model(spec.target_ref)
        sd = provider.get_state_dict(model)

        if name not in sd:
            raise DeleteTransformError(f"delete target disappeared during apply: {model}::{name}")

        del sd[name]










register_transform(DeleteTransform())



from __future__ import annotations

import re
from pathlib import Path
from typing import Any

import typer
from omegaconf import OmegaConf

_PATH_TOKEN_RE = re.compile(
    r"""
    ([^. \[\]]+)   # plain key
    |
    \[(\d+)\]      # numeric list index
    """,
    re.VERBOSE,
)


def is_yaml_file_arg(token: str) -> bool:
    path = Path(token)
    return path.suffix.lower() in {".yaml", ".yml"} and path.is_file()


def is_yaml_like_arg(token: str) -> bool:
    return Path(token).suffix.lower() in {".yaml", ".yml"}


def parse_override_path(path: str) -> list[str | int]:
    if not path:
        raise ValueError("override key must not be empty")

    tokens: list[str | int] = []
    pos = 0

    while pos < len(path):
        match = _PATH_TOKEN_RE.match(path, pos)
        if match is None:
            raise ValueError(f"invalid override path syntax: {path!r}")

        key, index = match.groups()
        if key is not None:
            tokens.append(key)
        else:
            tokens.append(int(index))

        pos = match.end()
        if pos < len(path):
            if path[pos] == ".":
                pos += 1
            elif path[pos] == "[":
                pass
            else:
                raise ValueError(f"invalid override path syntax: {path!r}")

    return tokens


def parse_override_value(raw_value: str) -> Any:
    try:
        parsed = OmegaConf.create({"_": raw_value})
        container = OmegaConf.to_container(parsed, resolve=True)
        if not isinstance(container, dict):
            raise ValueError("parsed override root is not a mapping")
        return container["_"]
    except Exception:
        return raw_value


def build_override_fragment(tokens: list[str | int], value: Any) -> Any:
    node: Any = value

    for token in reversed(tokens):
        if isinstance(token, int):
            arr = [None] * (token + 1)
            arr[token] = node
            node = arr
        else:
            node = {token: node}

    return node


def deep_merge_lists(base: list[Any], patch: list[Any]) -> list[Any]:
    size = max(len(base), len(patch))
    out: list[Any] = []

    for i in range(size):
        has_base = i < len(base)
        has_patch = i < len(patch)

        if has_patch and patch[i] is not None:
            if has_base and isinstance(base[i], dict) and isinstance(patch[i], dict):
                out.append(deep_merge_dicts(base[i], patch[i]))
            elif has_base and isinstance(base[i], list) and isinstance(patch[i], list):
                out.append(deep_merge_lists(base[i], patch[i]))
            else:
                out.append(patch[i])
        elif has_base:
            out.append(base[i])
        else:
            out.append(None)

    return out


def deep_merge_dicts(base: dict[str, Any], patch: dict[str, Any]) -> dict[str, Any]:
    out = dict(base)

    for key, patch_value in patch.items():
        if key not in out:
            out[key] = patch_value
            continue

        base_value = out[key]
        if isinstance(base_value, dict) and isinstance(patch_value, dict):
            out[key] = deep_merge_dicts(base_value, patch_value)
        elif isinstance(base_value, list) and isinstance(patch_value, list):
            out[key] = deep_merge_lists(base_value, patch_value)
        else:
            out[key] = patch_value

    return out


def apply_override(config: dict[str, Any], token: str) -> dict[str, Any]:
    if "=" not in token:
        raise ValueError("override must have the form key=value")

    raw_path, raw_value = token.split("=", 1)
    tokens = parse_override_path(raw_path)
    value = parse_override_value(raw_value)
    fragment = build_override_fragment(tokens, value)

    if not isinstance(fragment, dict):
        raise ValueError(f"top-level override must produce a mapping: {token!r}")

    return deep_merge_dicts(config, fragment)


def load_cli_config(tokens: list[str]) -> dict[str, Any]:
    merged: dict[str, Any] = {}

    for token in tokens:
        if is_yaml_file_arg(token):
            loaded = OmegaConf.to_container(OmegaConf.load(token), resolve=True)
            if loaded is None:
                loaded = {}
            if not isinstance(loaded, dict):
                raise typer.BadParameter(f"YAML file must contain a mapping: {token}")
            merged = deep_merge_dicts(merged, loaded)
            continue

        if is_yaml_like_arg(token):
            raise typer.BadParameter(f"YAML file does not exist: {token}")

        try:
            merged = apply_override(merged, token)
        except Exception as exc:
            raise typer.BadParameter(f"Invalid override {token!r}: {exc}") from exc

    return merged



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ..core import Expression, TransformError, collect_expr_models, compile_assert_expr, register_assert_expr
from ..core import StateDictProvider


@dataclass(frozen=True)
class AnyExpr:
    exprs: list[Expression]

    def evaluate(self, provider: StateDictProvider) -> None:
        errors: list[str] = []
        for expr in self.exprs:
            try:
                expr.evaluate(provider)
                return
            except TransformError as exc:
                errors.append(str(exc))
        raise TransformError("any failed: all alternatives failed:\n- " + "\n- ".join(errors))

    def collect_models(self) -> set[str]:
        return collect_expr_models(self.exprs)


@register_assert_expr(
    "any",
    payload_kind="list[assert-expr]",
    description="Succeeds if any inner assertion succeeds.",
)
def compile_any_expr(payload: Any, default_model: str | None) -> AnyExpr:
    if not isinstance(payload, list) or not payload:
        raise TransformError("any must be a non-empty list")
    return AnyExpr(exprs=[compile_assert_expr(item, default_model) for item in payload])



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from ..core import StateDictLike, StateDictProvider, TransformError


def _is_base_provider_instance(provider: object) -> bool:
    try:
        from .providers import BaseStateDictProvider
    except Exception:
        return False
    return isinstance(provider, BaseStateDictProvider)


def iter_alias_mappings(provider: StateDictProvider) -> list[tuple[str, dict[str, object]]]:
    mappings: list[tuple[str, dict[str, object]]] = []
    for attr_name in ("model_paths", "state_dicts", "_state_dicts"):
        value = getattr(provider, attr_name, None)
        if isinstance(value, dict):
            mappings.append((attr_name, value))
    return mappings


def list_model_aliases(provider: StateDictProvider | None) -> set[str]:
    if provider is None:
        return set()

    if _is_base_provider_instance(provider):
        return provider.list_model_aliases()

    list_aliases = getattr(provider, "list_model_aliases", None)
    if callable(list_aliases):
        aliases = list_aliases()
        if isinstance(aliases, set):
            return {str(alias) for alias in aliases}
        return {str(alias) for alias in aliases}

    aliases: set[str] = set()
    for _, mapping in iter_alias_mappings(provider):
        aliases.update(str(alias) for alias in mapping.keys())
    return aliases


def _has_model_alias(provider: StateDictProvider, alias: str) -> bool:
    if _is_base_provider_instance(provider):
        return provider.has_model_alias(alias)
    return alias in list_model_aliases(provider)


def get_or_create_alias_state_dict(
    provider: StateDictProvider,
    alias: str,
    *,
    error_type: type[TransformError],
    op_name: str,
) -> StateDictLike:
    if _is_base_provider_instance(provider):
        return provider.get_or_create_alias_state_dict(alias)
    if _has_model_alias(provider, alias):
        return provider.get_state_dict(alias)
    raise error_type(f"{op_name} requires a provider that supports creating new aliases")


def _list_loaded_tensor_names(provider: StateDictProvider | None) -> dict[str, set[str]]:
    if provider is None:
        return {}

    if _is_base_provider_instance(provider):
        items: Mapping[str, StateDictLike] = provider.state_dicts
    else:
        state_dicts = getattr(provider, "state_dicts", None)
        if not isinstance(state_dicts, dict):
            return {}
        items = state_dicts

    loaded: dict[str, set[str]] = {}
    for alias, state_dict in items.items():
        keys = getattr(state_dict, "keys", None)
        if not callable(keys):
            continue
        try:
            loaded[str(alias)] = {str(name) for name in keys()}
        except Exception:
            continue
    return loaded


def resolve_single_model_alias(
    provider: StateDictProvider,
    *,
    error_type: type[TransformError],
    op_name: str,
) -> str:
    aliases = list_model_aliases(provider)
    if len(aliases) != 1:
        raise error_type(f"{op_name}.alias is required when more than one model alias is available")
    return next(iter(aliases))


def find_alias_mapping(
    provider: StateDictProvider,
    alias: str,
    *,
    error_type: type[TransformError],
) -> tuple[str, dict[str, object], object]:
    for attr_name, mapping in iter_alias_mappings(provider):
        if alias in mapping:
            return attr_name, mapping, mapping[alias]
    raise error_type(f"unknown model prefix: {alias!r}")


def new_empty_state_dict(mappings: list[tuple[str, dict[str, object]]]) -> object:
    for _, mapping in mappings:
        for value in mapping.values():
            state_dict_type = type(value)
            try:
                return state_dict_type()
            except Exception:
                continue
    return {}


__all__ = [
    "iter_alias_mappings",
    "list_model_aliases",
    "get_or_create_alias_state_dict",
    "resolve_single_model_alias",
    "find_alias_mapping",
    "new_empty_state_dict",
]



from __future__ import annotations

from dataclasses import dataclass

from ..core import BinaryMappingSpec, DestinationPolicy
from ..core import ResolvedMapping, StateDictProvider, TensorRef, TransformError, select_tensor
from ..core import register_transform
from ..core import BinaryRefs, DeclarativeBinaryTransform, Docs


@dataclass(frozen=True)
class PermuteSpec(BinaryMappingSpec):
    order: tuple[int, ...]


def _build_permute_spec(
    from_ref: TensorRef, to_ref: TensorRef, payload: dict
) -> PermuteSpec:
    order = _parse_order(payload.get("order"))
    return PermuteSpec(from_ref=from_ref, to_ref=to_ref, order=order)


def _permute_apply(
    spec: PermuteSpec, item: ResolvedMapping, provider: StateDictProvider
) -> None:
    src_sd = provider.get_state_dict(item.src_model)
    dst_sd = provider.get_state_dict(item.dst_model)
    src_view = select_tensor(src_sd[item.src_name], item.src_slice)
    order = spec.order
    if src_view.dim() != len(order):
        raise TransformError(
            f"permute.order rank mismatch for {item.src_name}: "
            f"tensor rank {src_view.dim()} vs order length {len(order)}"
        )
    if sorted(order) != list(range(src_view.dim())):
        raise TransformError(
            f"permute.order must be a permutation of [0..{src_view.dim() - 1}], got {list(order)}"
        )
    dst_sd[item.dst_name] = src_view.permute(*order).clone()


class PermuteTransform(DeclarativeBinaryTransform[PermuteSpec]):
    name = "permute"
    error_type = TransformError
    spec_type = PermuteSpec
    destination_policy = DestinationPolicy.MUST_NOT_EXIST
    allowed_keys = {"from", "to", "order"}
    required_keys = {"from", "to", "order"}
    docs = Docs(
        "Permutes source tensor dimensions into new destination tensors.",
        notes=("'order' must be a full permutation of source dimensions.",),
        examples=("permute: { from: x, to: x_nhwc, order: [0, 2, 3, 1] }",),
    )
    refs = BinaryRefs(from_slice=True)
    spec_builder = staticmethod(_build_permute_spec)
    apply_fn = staticmethod(_permute_apply)


def _parse_order(raw: object) -> tuple[int, ...]:
    if not isinstance(raw, list) or not raw:
        raise TransformError("permute.order must be a non-empty list of integers")
    if not all(isinstance(x, int) for x in raw):
        raise TransformError("permute.order must be a non-empty list of integers")
    return tuple(raw)


register_transform(PermuteTransform())



from __future__ import annotations

import logging
from pathlib import Path
from typing import Dict

from .arena import ProviderError, SegmentedFileBackedArena
from .checkpoint_io import load_state_dict_from_path, persist_state_dict
from .output_model import infer_output_model
from .output_paths import parse_shard_size, resolve_output_destination
from .plan import SurgeryPlan
from .state_dicts import ArenaStateDict, InMemoryStateDict
from ..core import StateDictLike


logger = logging.getLogger("brainsurgery")


class BaseStateDictProvider:
    def __init__(self, model_paths: Dict[str, Path], max_io_workers: int):
        self.model_paths = model_paths
        self.max_io_workers = max_io_workers
        self.state_dicts: Dict[str, StateDictLike] = {}

    def get_state_dict(self, model: str) -> StateDictLike:
        raise NotImplementedError

    def create_state_dict(self) -> StateDictLike:
        raise NotImplementedError

    def list_model_aliases(self) -> set[str]:
        return set(self.model_paths) | set(self.state_dicts)

    def has_model_alias(self, model: str) -> bool:
        return model in self.list_model_aliases()

    def attach_state_dict(self, model: str, state_dict: StateDictLike) -> None:
        self.state_dicts[model] = state_dict
        self.model_paths.pop(model, None)

    def load_state_dict_from_checkpoint_path(self, path: Path) -> StateDictLike:
        state_dict = self.create_state_dict()
        load_state_dict_from_path(path, state_dict, max_io_workers=self.max_io_workers)
        return state_dict

    def load_alias_from_path(self, model: str, path: Path) -> StateDictLike:
        if self.has_model_alias(model):
            raise ProviderError(f"model alias already exists: {model!r}")
        state_dict = self.load_state_dict_from_checkpoint_path(path)
        self.attach_state_dict(model, state_dict)
        return state_dict

    def get_or_create_alias_state_dict(self, model: str) -> StateDictLike:
        if self.has_model_alias(model):
            return self.get_state_dict(model)
        state_dict = self.create_state_dict()
        self.attach_state_dict(model, state_dict)
        return state_dict

    def _get_or_load_state_dict(
        self,
        model: str,
        *,
        loaded_log_message: str,
    ) -> StateDictLike:
        if model in self.state_dicts:
            return self.state_dicts[model]

        if model not in self.model_paths:
            raise ProviderError(f"unknown model alias: {model!r}")

        path = self.model_paths[model]
        logger.info("Opening cranium for brain '%s' at %s", model, path)

        sd = self.load_state_dict_from_checkpoint_path(path)
        self.state_dicts[model] = sd
        logger.info(loaded_log_message, model, len(sd))

        return sd

    def close(self) -> None:
        pass

    def save_output(
        self,
        plan: SurgeryPlan,
        *,
        default_shard_size: str,
        max_io_workers: int,
    ) -> Path:
        if plan.output is None:
            raise ProviderError("save_output requires plan.output")

        output_model = infer_output_model(plan, self)
        state_dict = self.get_state_dict(output_model)

        output_path, output_format, shard_size = resolve_output_destination(
            plan.output,
            default_shard_size=default_shard_size,
        )

        logger.info(
            "Closing incision and preserving brain '%s' to %s (%s)",
            output_model,
            output_path,
            output_format,
        )

        written_path = persist_state_dict(
            dict(state_dict.items()),
            output_path=output_path,
            output_format=output_format,
            shard_size=shard_size,
            sharded_output_root=plan.output.path,
            max_io_workers=max_io_workers,
        )
        if shard_size is None:
            logger.info("Patient stable. Preserved %d tensors at %s", len(state_dict), written_path)
        else:
            logger.info(
                "Patient stable. Wrote %d tensors across sharded safetensors in %s",
                len(state_dict),
                written_path,
            )
        return written_path


class InMemoryStateDictProvider(BaseStateDictProvider):
    def get_state_dict(self, model: str) -> InMemoryStateDict:
        state_dict = self._get_or_load_state_dict(
            model,
            loaded_log_message="Brain '%s' exposed: %d tensors laid out on the operating table",
        )
        assert isinstance(state_dict, InMemoryStateDict)
        return state_dict

    def create_state_dict(self) -> InMemoryStateDict:
        return InMemoryStateDict()


class ArenaStateDictProvider(BaseStateDictProvider):
    def __init__(
        self,
        model_paths: Dict[str, Path],
        *,
        arena: SegmentedFileBackedArena,
        max_io_workers: int,
    ):
        super().__init__(model_paths, max_io_workers=max_io_workers)
        self.arena = arena

    def close(self) -> None:
        self.arena.close()

    def get_state_dict(self, model: str) -> ArenaStateDict:
        state_dict = self._get_or_load_state_dict(
            model,
            loaded_log_message=(
                "Brain '%s' transferred to surgical arena: %d tensors laid out on the operating table"
            ),
        )
        assert isinstance(state_dict, ArenaStateDict)
        return state_dict

    def create_state_dict(self) -> ArenaStateDict:
        return ArenaStateDict(self.arena)


def create_state_dict_provider(
    *,
    provider: str,
    model_paths: Dict[str, Path],
    max_io_workers: int,
    arena_root: Path,
    arena_segment_size: str,
) -> BaseStateDictProvider:
    provider_name = provider.strip().lower()

    if provider_name == "inmemory":
        return InMemoryStateDictProvider(
            model_paths,
            max_io_workers=max_io_workers,
        )

    if provider_name == "arena":
        segment_size_bytes = parse_shard_size(arena_segment_size)
        if segment_size_bytes is None:
            raise ProviderError("arena-segment-size must not be 'none'")

        arena = SegmentedFileBackedArena(
            arena_root,
            segment_size_bytes=segment_size_bytes,
        )
        return ArenaStateDictProvider(
            model_paths,
            arena=arena,
            max_io_workers=max_io_workers,
        )

    raise ProviderError("provider must be either 'inmemory' or 'arena'")



from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

import typer

from .config import load_cli_config
from .history import configure_history
from .interactive import normalize_transform_specs, prompt_interactive_transform
from .summary import build_raw_plan, write_executed_plan_summary
from ..engine import (
    compile_plan,
    execute_transform_pairs,
    ProviderError,
    create_state_dict_provider,
    list_model_aliases,
    use_output_emitter,
)

LOG_FORMAT = "%(asctime)s | %(levelname)-8s | %(message)s"
logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
logger = logging.getLogger("brainsurgery")
_ALLOWED_LOG_LEVELS = {"debug", "info", "warning", "error", "critical"}

app = typer.Typer(help="Brain surgery CLI.")


def configure_logging(log_level: str) -> None:
    level_name = log_level.strip().lower()
    if level_name not in _ALLOWED_LOG_LEVELS:
        raise typer.BadParameter(
            f"log-level must be one of: {', '.join(sorted(_ALLOWED_LOG_LEVELS))}"
        )

    level = getattr(logging, level_name.upper())
    logging.getLogger().setLevel(level)
    logger.setLevel(level)


@app.command()
def run(
    config_items: list[str] = typer.Argument(
        None,
        help=(
            "YAML plan fragments and/or key=value overrides. YAML files are loaded and "
            "deep-merged in order; overrides are applied last. Typically used to define "
            "inputs, transforms, and output."
        ),
    ),
    shard_size: str = typer.Option(
        "5GB",
        help="Default shard size when writing directory outputs (e.g. safetensors shards). Ignored for single-file outputs.",
    ),
    num_workers: int = typer.Option(
        8,
        help="Maximum parallel workers for loading and saving tensors. Higher values improve I/O throughput but increase memory pressure.",
    ),
    provider: str = typer.Option(
        "inmemory",
        help="State-dict backend. 'inmemory' loads tensors into RAM; 'arena' memory-maps tensors to disk for large models.",
    ),
    arena_root: Path = typer.Option(
        Path(".brainsurgery"),
        help="Directory for arena storage when using --provider arena (memory-mapped tensor backing).",
    ),
    arena_segment_size: str = typer.Option(
        "1GB",
        help="Segment size for arena storage (e.g. 1GB). Larger segments reduce fragmentation but use more disk.",
    ),
    interactive: bool = typer.Option(
        False,
        "-i",
        "--interactive",
        help="Run configured transforms, then enter an interactive prompt to execute additional transforms incrementally.",
    ),
    summarize: bool = typer.Option(
        True,
        "-s",
        "--summarize/--no-summarize",
        help="Write a YAML summary of the transforms actually executed (after overrides and interactive edits).",
    ),
    summarize_path: Path | None = typer.Option(
        None,
        help="Destination for the executed-plan summary. Defaults to stdout if not set.",
    ),
    log_level: str = typer.Option(
        "info",
        "--log-level",
        help="Logging verbosity (debug, info, warning, error, critical).",
    ),
) -> None:
    """Load a plan, execute it, and save the rewritten output checkpoint."""
    configure_logging(log_level)
    configure_history()

    raw_plan = load_cli_config(config_items or [])

    logger.info(
        "Scrubbing in. Surgical plan assembled from %d config item(s)",
        len(config_items or []),
    )
    surgery_plan = compile_plan(raw_plan)
    logger.info(
        "Surgical plan ready: %d brain(s) prepped, %d procedure(s) scheduled, preservation %s",
        len(surgery_plan.inputs),
        len(surgery_plan.transforms),
        surgery_plan.output.path if surgery_plan.output else None,
    )

    try:
        state_dict_provider = create_state_dict_provider(
            provider=provider,
            model_paths=surgery_plan.inputs,
            max_io_workers=num_workers,
            arena_root=arena_root,
            arena_segment_size=arena_segment_size,
        )
    except ProviderError as exc:
        raise typer.BadParameter(str(exc)) from exc

    executed_transforms: list[dict[str, Any]] = []
    written_path: str | Path | None = None

    try:
        with use_output_emitter(typer.echo):
            configured_pairs = zip(
                normalize_transform_specs(raw_plan.get("transforms")),
                surgery_plan.transforms,
                strict=False,
            )
            should_continue, newly_executed = execute_transform_pairs(
                configured_pairs,
                state_dict_provider,
                interactive=False,
            )
            executed_transforms.extend(newly_executed)

            if should_continue and interactive:
                logger.info("Entering interactive mode after configured procedures")

                while True:
                    extra_specs = prompt_interactive_transform(state_dict_provider=state_dict_provider)
                    if extra_specs is None:
                        logger.info("Interactive session complete")
                        break

                    interactive_inputs = raw_plan.get("inputs", [])
                    aliases = sorted(list_model_aliases(state_dict_provider))
                    if aliases:
                        interactive_inputs = [f"{alias}::/dev/null" for alias in aliases]

                    interactive_raw_plan = build_raw_plan(
                        inputs=interactive_inputs,
                        output=raw_plan.get("output"),
                        transforms=extra_specs,
                    )

                    try:
                        interactive_plan = compile_plan(interactive_raw_plan)
                    except Exception as exc:
                        logger.error("Could not compile interactive transform(s): %s", exc)
                        continue

                    interactive_pairs = zip(
                        extra_specs,
                        interactive_plan.transforms,
                        strict=False,
                    )
                    should_continue, newly_executed = execute_transform_pairs(
                        interactive_pairs,
                        state_dict_provider,
                        interactive=True,
                    )
                    executed_transforms.extend(newly_executed)

                    if not should_continue:
                        logger.info("Leaving interactive mode")
                        break

            if surgery_plan.output is None:
                logger.info("No preservation requested; concluding operation without closure")
            else:
                written_path = state_dict_provider.save_output(
                    surgery_plan,
                    default_shard_size=shard_size,
                    max_io_workers=num_workers,
                )
                logger.info("Operation complete. Brain preserved at %s", written_path)

            if summarize:
                write_executed_plan_summary(
                    inputs=raw_plan.get("inputs", []),
                    output=raw_plan.get("output"),
                    transforms=executed_transforms,
                    destination=summarize_path,
                )

    finally:
        state_dict_provider.close()


if __name__ == "__main__":
    app()


__all__ = ["app", "configure_logging", "run", "logger"]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

from ..engine.tensor_files import load_tensor_from_path
from ..engine import get_or_create_alias_state_dict
from ..engine import BaseStateDictProvider, ProviderError
from ..core import parse_model_expr
from ..core import StateDictProvider, TransformError
from ..core import TypedTransform, TransformResult, register_transform
from ..core import ensure_mapping_payload, require_nonempty_string, validate_payload_keys


class LoadTransformError(TransformError):
    pass


@dataclass(frozen=True)
class LoadSpec:
    path: Path
    alias: str
    tensor_name: str | None
    format: str

    def collect_models(self) -> set[str]:
        # load can introduce a new model alias
        return set()


class LoadTransform(TypedTransform[LoadSpec]):
    name = "load"
    error_type = LoadTransformError
    spec_type = LoadSpec
    allowed_keys = {"path", "alias", "to", "format"}
    required_keys = {"path"}
    help_text = (
        "Loads either a full state_dict or a single tensor from disk.\n"
        "\n"
        "Without 'to', loads a full state_dict into 'alias'. With 'to', loads one tensor "
        "into a tensor name (optionally with alias in 'to', e.g. model::name).\n"
        "\n"
        "Examples:\n"
        "  load: { path: /tmp/a.safetensors, alias: a }\n"
        "  load: { path: /tmp/tensor.npy, to: model::embed.weight }\n"
        "  load: /tmp/model.safetensors"
    )

    def completion_reference_keys(self) -> list[str]:
        return ["to"]

    def compile(self, payload: Any, default_model: str | None) -> LoadSpec:
        if isinstance(payload, str):
            payload = {"path": payload}
        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
            required_keys=self.required_keys,
        )

        path = Path(require_nonempty_string(payload, op_name=self.name, key="path"))
        raw_alias = payload.get("alias")
        if raw_alias is not None and (not isinstance(raw_alias, str) or not raw_alias):
            raise LoadTransformError("load.alias must be a non-empty string when provided")

        raw_format = payload.get("format", "auto")
        if not isinstance(raw_format, str) or not raw_format:
            raise LoadTransformError("load.format must be a non-empty string when provided")
        fmt = raw_format.strip().lower()
        if fmt not in {"auto", "torch", "safetensors", "numpy"}:
            raise LoadTransformError("load.format must be one of: auto, torch, safetensors, numpy")

        alias_default = raw_alias or default_model or "model"
        tensor_name: str | None = None
        raw_to = payload.get("to")
        if raw_to is not None:
            if not isinstance(raw_to, str) or not raw_to:
                raise LoadTransformError("load.to must be a non-empty string when provided")
            target_ref = parse_model_expr(raw_to, default_model=alias_default)
            if target_ref.slice_spec is not None:
                raise LoadTransformError("load.to must not be sliced")
            if not isinstance(target_ref.expr, str):
                raise LoadTransformError("load.to must resolve to a single tensor name")
            if raw_alias is not None and target_ref.model != raw_alias:
                raise LoadTransformError("load.alias conflicts with model alias in load.to")
            assert target_ref.model is not None
            alias_default = target_ref.model
            tensor_name = target_ref.expr
        elif fmt != "auto":
            raise LoadTransformError("load.format is only supported for tensor loads (with load.to)")

        return LoadSpec(path=path, alias=alias_default, tensor_name=tensor_name, format=fmt)

    def apply(self, spec: object, provider: StateDictProvider) -> TransformResult:
        typed = self.require_spec(spec)

        if typed.tensor_name is None:
            if not isinstance(provider, BaseStateDictProvider):
                raise LoadTransformError("load requires a provider that supports creating new aliases")
            try:
                state_dict = provider.load_alias_from_path(typed.alias, typed.path)
            except ProviderError as exc:
                message = str(exc).replace("model alias", "load alias")
                raise LoadTransformError(message) from exc
            except RuntimeError as exc:
                raise LoadTransformError(str(exc)) from exc
            return TransformResult(name=self.name, count=len(state_dict))

        try:
            tensor = load_tensor_from_path(typed.path, format=typed.format)  # type: ignore[arg-type]
        except RuntimeError as exc:
            raise LoadTransformError(str(exc)) from exc
        state_dict = get_or_create_alias_state_dict(
            provider,
            typed.alias,
            error_type=LoadTransformError,
            op_name=self.name,
        )
        if typed.tensor_name in state_dict:
            raise LoadTransformError(
                f"load destination already exists: {typed.alias}::{typed.tensor_name}"
            )
        state_dict[typed.tensor_name] = tensor
        return TransformResult(name=self.name, count=1)

    def infer_output_model(self, spec: object) -> str:
        return self.require_spec(spec).alias










register_transform(LoadTransform())



from __future__ import annotations

from ..core import (
    BinaryMappingSpec,
    BinaryRefs,
    DeclarativeBinaryTransform,
    DestinationPolicy,
    Docs,
    ResolvedMapping,
    ResolvedTernaryMapping,
    StateDictProvider,
    TernaryMappingSpec,
    TransformError,
    require_same_shape_dtype_device,
    select_tensor,
)
from ..core import require_same_shape_dtype_device3
from ..core import register_transform
from ..core import DeclarativeTernaryTransform, TernaryRefs


def _subtract_apply(
    _spec: TernaryMappingSpec,
    item: ResolvedTernaryMapping,
    provider: StateDictProvider,
) -> None:
    src_a_sd = provider.get_state_dict(item.src_a_model)
    src_b_sd = provider.get_state_dict(item.src_b_model)
    dst_sd = provider.get_state_dict(item.dst_model)

    src_a_view = select_tensor(src_a_sd[item.src_a_name], item.src_a_slice)
    src_b_view = select_tensor(src_b_sd[item.src_b_name], item.src_b_slice)
    dst_view = select_tensor(dst_sd[item.dst_name], item.dst_slice)

    require_same_shape_dtype_device3(
        src_a_view,
        src_b_view,
        dst_view,
        op_name="subtracting",
        first_name=item.src_a_name,
        second_name=item.src_b_name,
        dest_name=item.dst_name,
        symbol="-",
    )
    dst_view.copy_(src_a_view - src_b_view)
    dst_sd.mark_write(item.dst_name)


class SubtractTransform(DeclarativeTernaryTransform[TernaryMappingSpec]):
    name = "subtract"
    error_type = TransformError
    destination_policy = DestinationPolicy.MUST_EXIST
    docs = Docs(
        "Computes elementwise subtraction from 'from_a' minus 'from_b' into 'to'.",
        examples=("subtract: { from_a: a.weight, from_b: b.weight, to: out.weight }",),
    )
    refs = TernaryRefs(from_a_slice=True, from_b_slice=True, to_slice=True)
    apply_fn = staticmethod(_subtract_apply)


register_transform(SubtractTransform())


def _subtract_in_place_apply(
    _spec: BinaryMappingSpec,
    item: ResolvedMapping,
    provider: StateDictProvider,
) -> None:
    src_sd = provider.get_state_dict(item.src_model)
    dst_sd = provider.get_state_dict(item.dst_model)

    src_view = select_tensor(src_sd[item.src_name], item.src_slice)
    dst_view = select_tensor(dst_sd[item.dst_name], item.dst_slice)
    require_same_shape_dtype_device(
        src_view,
        dst_view,
        op_name="subtracting",
        left_name=item.src_name,
        right_name=item.dst_name,
    )

    dst_view.sub_(src_view)
    dst_sd.mark_write(item.dst_name)


class SubtractInPlaceTransform(DeclarativeBinaryTransform[BinaryMappingSpec]):
    name = "subtract_"
    error_type = TransformError
    destination_policy = DestinationPolicy.MUST_EXIST
    docs = Docs(
        "Subtracts source tensors from destination tensors in-place.",
        notes=("Computes: to <- to - from.",),
        examples=(
            "subtract_: { from: delta.weight, to: model.weight }",
            "subtract_: { from: 'a::[:, :10]', to: 'b::[:, :10]' }",
        ),
    )
    refs = BinaryRefs(from_slice=True, to_slice=True)
    apply_fn = staticmethod(_subtract_in_place_apply)


register_transform(SubtractInPlaceTransform())



from __future__ import annotations

from ..core import BinaryMappingSpec, BinaryMappingTransform, DestinationPolicy
from ..core import ResolvedMapping
from ..core import TensorRef, parse_slice, select_tensor
from ..core import require_same_shape_dtype_device
from ..core import register_transform
from ..core import StateDictProvider, TransformError


class AssignTransform(BinaryMappingTransform[BinaryMappingSpec]):
    name = "assign"
    spec_type = BinaryMappingSpec
    destination_policy = DestinationPolicy.MUST_EXIST
    progress_desc = "Applying assign transforms"
    help_text = (
        "Copies tensor values from 'from' into 'to'. The destination tensor must "
        "already exist. Source and destination (after slicing) must have identical shapes.\n"
        "\n"
        "Both references support slicing.\n"
        "\n"
        "Examples:\n"
        "  assign: { from: a.weight, to: b.weight }\n"
        "  assign: { from: 'a.weight::[:, :10]', to: 'b.weight::[:, :10]' }"
    )

    def validate_refs(self, from_ref: TensorRef, to_ref: TensorRef) -> None:
        if from_ref.slice_spec is not None:
            parse_slice(from_ref.slice_spec)
        if to_ref.slice_spec is not None:
            parse_slice(to_ref.slice_spec)

    def apply_mapping(self, item: ResolvedMapping, provider: StateDictProvider) -> None:
        src_sd = provider.get_state_dict(item.src_model)
        dst_sd = provider.get_state_dict(item.dst_model)

        src_view = select_tensor(src_sd[item.src_name], item.src_slice)
        dst_view = select_tensor(dst_sd[item.dst_name], item.dst_slice)
        require_same_shape_dtype_device(
            src_view,
            dst_view,
            op_name="assigning",
            left_name=item.src_name,
            right_name=item.dst_name,
        )

        dst_view.copy_(src_view)
        dst_sd.mark_write(item.dst_name)










register_transform(AssignTransform())



from __future__ import annotations

from dataclasses import dataclass

from ..core import ResolvedMapping, StateDictProvider, TensorRef, TransformError, must_model, parse_slice, select_tensor
from ..core import register_transform
from ..core import require_numeric
from ..core import BinaryMappingSpec, DestinationPolicy
from ..core import BinaryRefs, DeclarativeBinaryTransform, Docs, UnaryRefs, UnarySpec, DeclarativeUnaryTransform


@dataclass(frozen=True)
class ScaleSpec(BinaryMappingSpec):
    factor: float


def _build_scale_spec(
    from_ref: TensorRef, to_ref: TensorRef, payload: dict
) -> ScaleSpec:
    factor = require_numeric(payload, op_name="scale", key="by")
    return ScaleSpec(from_ref=from_ref, to_ref=to_ref, factor=factor)


def _scale_apply(
    spec: ScaleSpec, item: ResolvedMapping, provider: StateDictProvider
) -> None:
    src_sd = provider.get_state_dict(item.src_model)
    dst_sd = provider.get_state_dict(item.dst_model)
    scaled = select_tensor(src_sd[item.src_name], item.src_slice).clone()
    scaled.mul_(spec.factor)
    dst_sd[item.dst_name] = scaled


class ScaleTransform(DeclarativeBinaryTransform[ScaleSpec]):
    name = "scale"
    error_type = TransformError
    spec_type = ScaleSpec
    destination_policy = DestinationPolicy.MUST_NOT_EXIST
    allowed_keys = {"from", "to", "by"}
    required_keys = {"from", "to", "by"}
    docs = Docs(
        "Scales source tensors by a numeric factor into new destination tensors.",
        examples=(
            "scale: { from: ln_f.weight, to: ln_f_half.weight, by: 0.5 }",
            "scale: { from: '.*bias', to: 'scaled.\\\\g<0>', by: -1 }",
            "scale: { from: 'h.0.attn.c_attn.weight::[:, :10]', to: h.0.attn.c_attn.partial, by: 2.0 }",
        ),
    )
    refs = BinaryRefs(from_slice=True)
    spec_builder = staticmethod(_build_scale_spec)
    apply_fn = staticmethod(_scale_apply)


register_transform(ScaleTransform())


@dataclass(frozen=True)
class ScaleInPlaceSpec(UnarySpec):
    factor: float


def _build_scale_in_place_spec(
    target_ref: TensorRef, payload: dict
) -> ScaleInPlaceSpec:
    factor = require_numeric(payload, op_name="scale_", key="by")
    return ScaleInPlaceSpec(target_ref=target_ref, factor=factor)


def _scale_in_place_apply(
    spec: ScaleInPlaceSpec, name: str, provider: StateDictProvider
) -> None:
    model = must_model(spec.target_ref)
    sd = provider.get_state_dict(model)
    tensor = sd[name]

    slice_spec = (
        parse_slice(spec.target_ref.slice_spec)
        if spec.target_ref.slice_spec is not None
        else None
    )
    view = select_tensor(tensor, slice_spec)
    view.mul_(spec.factor)
    sd.mark_write(name)


class ScaleInPlaceTransform(DeclarativeUnaryTransform[ScaleInPlaceSpec]):
    name = "scale_"
    error_type = TransformError
    spec_type = ScaleInPlaceSpec
    allowed_keys = {"target", "by"}
    required_keys = {"target", "by"}
    docs = Docs(
        "Scales tensors in-place by a numeric factor.",
        notes=("The selected tensor (or slice) is multiplied by 'by'.",),
        examples=(
            "scale_: { target: ln_f.weight, by: 0.5 }",
            "scale_: { target: '.*bias', by: -1 }",
            "scale_: { target: 'h.0.attn.c_attn.weight::[:, :10]', by: 2.0 }",
        ),
    )
    refs = UnaryRefs(target_slice=True)
    spec_builder = staticmethod(_build_scale_in_place_spec)
    apply_fn = staticmethod(_scale_in_place_apply)


register_transform(ScaleInPlaceTransform())



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ..core import Expression, TransformError, compile_assert_expr, register_assert_expr
from ..core import StateDictProvider


@dataclass(frozen=True)
class NotExpr:
    expr: Expression

    def evaluate(self, provider: StateDictProvider) -> None:
        try:
            self.expr.evaluate(provider)
        except TransformError:
            return

        raise TransformError(f"not failed: inner assertion succeeded: {self.expr!r}")

    def collect_models(self) -> set[str]:
        return self.expr.collect_models()


@register_assert_expr(
    "not",
    payload_kind="assert-expr",
    description="Succeeds if the inner assertion fails.",
)
def compile_not_expr(payload: Any, default_model: str | None) -> Expression:
    return NotExpr(expr=compile_assert_expr(payload, default_model))



from __future__ import annotations

from dataclasses import dataclass
import re
from typing import Any, Literal

from ..engine import (
    find_alias_mapping,
    get_or_create_alias_state_dict,
    iter_alias_mappings,
    list_model_aliases,
    new_empty_state_dict,
)
from ..engine.frontend import emit_line
from ..core import TransformError
from ..core import TypedTransform, TransformResult, register_transform
from ..core import ensure_mapping_payload, require_nonempty_string, validate_payload_keys
from ..core import StateDictProvider


class PrefixesTransformError(TransformError):
    pass


PrefixesMode = Literal["list", "add", "remove", "rename"]


@dataclass(frozen=True)
class PrefixesSpec:
    mode: PrefixesMode
    alias: str | None = None
    source_alias: str | None = None
    dest_alias: str | None = None

    def collect_models(self) -> set[str]:
        return set()


class PrefixesTransform(TypedTransform[PrefixesSpec]):
    name = "prefixes"
    error_type = PrefixesTransformError
    spec_type = PrefixesSpec
    completion_requires_payload = False
    allowed_keys = {"mode", "alias", "from", "to"}
    help_text = (
        "Lists or edits the currently available model prefixes (aliases).\n"
        "\n"
        "Modes:\n"
        "  - list (default): show all available aliases\n"
        "  - add: create a new empty alias (`alias`)\n"
        "  - remove: delete an existing alias (`alias`)\n"
        "  - rename: rename an existing alias (`from`, `to`)\n"
        "\n"
        "Examples:\n"
        "  prefixes\n"
        "  prefixes: { mode: list }\n"
        "  prefixes: { mode: add, alias: scratch }\n"
        "  prefixes: { mode: remove, alias: scratch }\n"
        "  prefixes: { mode: rename, from: scratch, to: edited }"
    )

    def compile(self, payload: Any, default_model: str | None) -> PrefixesSpec:
        del default_model

        if payload is None:
            return PrefixesSpec(mode="list")

        if isinstance(payload, dict) and not payload:
            return PrefixesSpec(mode="list")

        payload = ensure_mapping_payload(payload, self.name)
        validate_payload_keys(
            payload,
            op_name=self.name,
            allowed_keys=self.allowed_keys,
        )

        raw_mode = payload.get("mode", "list")
        if not isinstance(raw_mode, str) or not raw_mode:
            raise PrefixesTransformError("prefixes.mode must be a non-empty string when provided")

        mode = raw_mode.strip().lower()
        if mode == "list":
            _require_only_keys(payload, allowed={"mode"})
            return PrefixesSpec(mode="list")

        if mode == "add":
            _require_only_keys(payload, allowed={"mode", "alias"})
            return PrefixesSpec(
                mode="add",
                alias=require_nonempty_string(payload, op_name=self.name, key="alias"),
            )

        if mode == "remove":
            _require_only_keys(payload, allowed={"mode", "alias"})
            return PrefixesSpec(
                mode="remove",
                alias=require_nonempty_string(payload, op_name=self.name, key="alias"),
            )

        if mode == "rename":
            _require_only_keys(payload, allowed={"mode", "from", "to"})
            return PrefixesSpec(
                mode="rename",
                source_alias=require_nonempty_string(payload, op_name=self.name, key="from"),
                dest_alias=require_nonempty_string(payload, op_name=self.name, key="to"),
            )

        raise PrefixesTransformError("prefixes.mode must be one of: list, add, remove, rename")

    def apply(self, spec: object, provider: StateDictProvider) -> TransformResult:
        typed = self.require_spec(spec)

        if typed.mode == "list":
            aliases = sorted(list_model_aliases(provider))
            if aliases:
                emit_line("Available model prefixes:")
                for alias in aliases:
                    emit_line(f"  {alias}::")
            else:
                emit_line("No model prefixes available.")
            return TransformResult(name=self.name, count=len(aliases))

        if typed.mode == "add":
            assert typed.alias is not None
            _create_empty_alias(provider, typed.alias)
            return TransformResult(name=self.name, count=1)

        if typed.mode == "remove":
            assert typed.alias is not None
            _delete_alias(provider, typed.alias)
            return TransformResult(name=self.name, count=1)

        if typed.mode == "rename":
            assert typed.source_alias is not None
            assert typed.dest_alias is not None
            _rename_alias(provider, source=typed.source_alias, dest=typed.dest_alias)
            return TransformResult(name=self.name, count=1)

        raise PrefixesTransformError(f"unsupported prefixes mode: {typed.mode}")

    def infer_output_model(self, spec: object) -> str:
        typed = self.require_spec(spec)
        if typed.mode == "add":
            assert typed.alias is not None
            return typed.alias
        if typed.mode == "rename":
            assert typed.dest_alias is not None
            return typed.dest_alias
        raise PrefixesTransformError("prefixes does not infer an output model in this mode")

    def contributes_output_model(self, spec: object) -> bool:
        typed = self.require_spec(spec)
        return typed.mode in {"add", "rename"}

    def completion_key_candidates(self, before_cursor: str, prefix_text: str) -> list[str] | None:
        used_keys = set(re.findall(r"([A-Za-z_][A-Za-z0-9_]*)\s*:", before_cursor))
        mode = _prefixes_mode(before_cursor)
        if mode is None:
            options = ["mode: "]
        elif mode == "list":
            options = []
        elif mode in {"add", "remove"}:
            options = ["alias: "]
        elif mode == "rename":
            options = ["from: ", "to: "]
        else:
            options = ["mode: "]
        filtered = [
            candidate
            for candidate in options
            if candidate[:-2] not in used_keys and candidate.startswith(prefix_text)
        ]
        if filtered:
            return filtered
        if options and not all(option[:-2] in used_keys for option in options):
            return []
        return ["}"]

    def completion_value_candidates(
        self,
        value_key: str | None,
        prefix_text: str,
        model_aliases: list[str],
    ) -> list[str] | None:
        if value_key == "mode":
            return [
                mode for mode in ("list", "add", "remove", "rename") if mode.startswith(prefix_text)
            ]
        if value_key in {"alias", "from", "to"}:
            return [alias for alias in sorted(model_aliases) if alias.startswith(prefix_text)]
        return None

    def completion_reference_keys(self) -> list[str]:
        return []


def _require_only_keys(payload: dict[str, object], *, allowed: set[str]) -> None:
    unexpected = set(payload) - allowed
    if unexpected:
        raise PrefixesTransformError(f"prefixes received unknown keys: {sorted(unexpected)}")


def _prefixes_mode(before_cursor: str) -> str | None:
    match = re.search(r"\bmode\s*:\s*([A-Za-z_][A-Za-z0-9_]*)", before_cursor)
    if match is None:
        return None
    return match.group(1).lower()


def _list_aliases(provider: StateDictProvider) -> set[str]:
    return list_model_aliases(provider)


def _iter_alias_mappings(provider: StateDictProvider) -> list[tuple[str, dict[str, object]]]:
    return iter_alias_mappings(provider)


def _find_alias_mapping(provider: StateDictProvider, alias: str) -> tuple[str, dict[str, object], object]:
    return find_alias_mapping(provider, alias, error_type=PrefixesTransformError)


def _create_empty_alias(provider: StateDictProvider, alias: str) -> None:
    if alias in _list_aliases(provider):
        raise PrefixesTransformError(f"model prefix already exists: {alias!r}")

    try:
        get_or_create_alias_state_dict(
            provider,
            alias,
            error_type=PrefixesTransformError,
            op_name="prefixes add",
        )
        return
    except PrefixesTransformError:
        pass

    mappings = _iter_alias_mappings(provider)
    if not mappings:
        raise PrefixesTransformError("prefixes add requires a provider that supports editable aliases")

    _, mapping = mappings[-1]
    mapping[alias] = _new_empty_state_dict(mappings)


def _new_empty_state_dict(mappings: list[tuple[str, dict[str, object]]]) -> object:
    return new_empty_state_dict(mappings)


def _delete_alias(provider: StateDictProvider, alias: str) -> None:
    _, mapping, _ = _find_alias_mapping(provider, alias)
    del mapping[alias]


def _rename_alias(provider: StateDictProvider, *, source: str, dest: str) -> None:
    if source == dest:
        raise PrefixesTransformError("prefixes rename source and destination must differ")
    if dest in _list_aliases(provider):
        raise PrefixesTransformError(f"model prefix already exists: {dest!r}")
    _, mapping, value = _find_alias_mapping(provider, source)
    del mapping[source]
    mapping[dest] = value






register_transform(PrefixesTransform())
